# -*-: coding utf-8 -*-
""" Thread handler. """

import threading
import time

from .singleton import Singleton
from .usb_utils import USB


class ThreadHandler(Singleton):
    """ Thread handler. """

    def __init__(self):
        """ Initialisation. """
        self.thread_pool = []
        self.run_events = []

    def run(self, target, args=()):
        """ Run a function in a separate thread.

        :param target: the function to run.
        :param args: the parameters to pass to the function.
        """
        run_event = threading.Event()
        run_event.set()
        thread = threading.Thread(target=target, args=args + (run_event, ))
        self.thread_pool.append(thread)
        self.run_events.append(run_event)
        thread.start()

    def start_run_loop(self):
        """ Start the thread handler, ensuring that everything stops property
            when sending a keyboard interrup.
        """
        try:
            while 1:
                time.sleep(.1)
        except KeyboardInterrupt:
            print("BREAK")
            self.stop()

    def stop(self):
        """ Stop all functions running in the thread handler."""
        for run_event in self.run_events:
            run_event.clear()

        for thread in self.thread_pool:
            thread.join()



# -*-: coding utf-8 -*-
""" Singleton class. """


class Singleton(object):
    """ Singleton class. """

    _instance = None

    def __new__(cls, *args, **kwargs):
        """ Initialisation. """
        if not cls._instance:
            cls._instance = super(Singleton, cls).__new__(cls)
        return cls._instance



# -*-: coding utf-8 -*-
""" Wrapper for various TTS services (currently only Google). """

import json
import os

from gtts import gTTS

from .logging import debug_log

from .audio_player import AudioPlayer


class TTS:
    """ Wrapper for TTS service. """

    class Provider:
        """ Supported providers. """
        google = range(1)

    def __init__(self, provider, locale):
        """ Initialise the service.

        :param provider: the TTS provider, e.g. TTS.Provider.google.
        :param locale: the short language locale, e.g. "fr".
        """
        self.tts = GTTS(locale)

    def speak(self, sentence):
        """ Speak a sentence.

        :param sentence: the sentence to speak.
        """
        if self.tts is None:
            return
        self.tts.speak(sentence)


class GTTS:
    """ Google TTS service. """

    def __init__(self, locale):
        """ Initialise the service.

        :param locale: the language locale, e.g. "fr" or "en_US".
        """
        self.locale = locale.split("_")[0]

    def speak(self, sentence):
        """ Speak a sentence using Google TTS.

        :param sentence: the sentence to speak.
        """
        snips_dir = ".snips"
        filename = "gtts.mp3"
        file_path = "{}/{}".format(snips_dir, filename)

        if not os.path.exists(snips_dir):
            os.makedirs(snips_dir)

        def delete_file():
            try:
                os.remove(file_path)
                if not os.listdir(snips_dir):
                    try:
                        os.rmdir(snips_dir)
                    except OSError:
                        pass
            except:
                pass

        debug_log("Speak: {}".format(sentence))
        tts = gTTS(text=sentence, lang=self.locale)
        tts.save(file_path)
        AudioPlayer.play_async(file_path, delete_file)



# -*-: coding utf-8 -*-
""" A representation of a datetime with a given granularity (day, week). """


# pylint: disable=too-few-public-methods
class InstantTime:
    """ A representation of a datetime with a given granularity (day, week).
    """

    day, week = range(2)

    def __init__(self, datetime, granularity=None):
        """ Initialisation.

        :param datetime: the underlying datetime object
        :param granularity: granularity of the datetime, either
                            InstantTime.day or InstantTime.week.
        """
        self.datetime = datetime
        self.granularity = granularity or InstantTime.day

    @staticmethod
    def parse_grain(grain):
        """ Parse a string to a granularity, e.g. "Day" to InstantTime.day.

        :param grain: a string representing a granularity.
        """
        if not grain:
            return InstantTime.day
        if grain.lower() == 'week':
            return InstantTime.week
        return InstantTime.day



# -*-: coding utf-8 -*-
""" A representation of a time interval. """


# pylint: disable=too-few-public-methods
class TimeInterval:
    """ A representation of a time interval. """

    def __init__(self, start, end):
        """ Initialisation.

        :param start: the start of the time interval.
        :param end: the end of the time interval.
        """
        self.start = start
        self.end = end



# -*-: coding utf-8 -*-
""" Sound service for playing various state sounds. """

import os
import sys

from .audio_player import AudioPlayer


MODULE_PATH = os.path.dirname(sys.modules[__name__].__file__)
REL_SOUND_DIR = "data/sounds"
ABS_SOUND_DIR = "{}/{}".format(MODULE_PATH, REL_SOUND_DIR)

class SoundService:
    """ Sound service for playing various state sounds. """

    class State:
        """ States handled by the sound service. """
        none, welcome, goodbye, hotword_detected, asr_text_captured, error = range(
            6)

    @staticmethod
    def play(state):
        """ Play sound for a given state.

        :param state: a State value.
        """
        filename = None
        if state == SoundService.State.welcome:
            filename = "pad_glow_welcome1.wav"
        elif state == SoundService.State.goodbye:
            filename = "pad_glow_power_off.wav"
        elif state == SoundService.State.hotword_detected:
            filename = "pad_soft_on.wav"
        elif state == SoundService.State.asr_text_captured:
            filename = "pad_soft_off.wav"
        elif state == SoundService.State.error:
            filename = "music_marimba_error_chord_2x.wav"

        if filename is not None:
            AudioPlayer.play_async("{}/{}".format(ABS_SOUND_DIR, filename))



# -*-: coding utf-8 -*-
""" Snips core server. """

import json
import time

from socket import error as socket_error

import paho.mqtt.client as mqtt

from .logging import debug_log, LOGGING_ENABLED
from .thread_handler import ThreadHandler
from .intent_parser import IntentParser
from .state_handler import StateHandler, State


class Server():
    """ Snips core server. """

    def __init__(self,
                 mqtt_hostname,
                 mqtt_port,
                 logging_enabled,
                 registry,
                 handle_intent):
        """ Initialisation.

        :param config: a YAML configuration.
        :param assistant: the client assistant class, holding the
                          intent handler and intents registry.
        """
        self.registry = registry
        self.handle_intent = handle_intent
        self.thread_handler = ThreadHandler()
        self.state_handler = StateHandler(self.thread_handler)

        self.client = mqtt.Client()
        self.client.on_connect = self.on_connect
        self.client.on_disconnect = self.on_disconnect
        self.client.on_message = self.on_message
        self.mqtt_hostname = mqtt_hostname
        self.mqtt_port = mqtt_port

        LOGGING_ENABLED = logging_enabled

        self.first_hotword_detected = False

    def start(self):
        """ Start the MQTT client. """
        self.thread_handler.run(target=self.start_blocking)
        self.thread_handler.start_run_loop()

    def start_blocking(self, run_event):
        """ Start the MQTT client, as a blocking method.

        :param run_event: a run event object provided by the thread handler.
        """
        topic = "#"
        debug_log("Connecting to {} on port {}".format(self.mqtt_hostname,
                                                       str(self.mqtt_port)))
        while True and run_event.is_set():
            try:
                debug_log("Trying to connect to {}".format(self.mqtt_hostname))
                self.client.connect(self.mqtt_hostname, self.mqtt_port, 60)
                break
            except socket_error:
                time.sleep(5)
            except Exception as e:
                time.sleep(5)
        self.client.subscribe(topic, 0)
        while run_event.is_set():
            self.client.loop()

    # pylint: disable=unused-argument,no-self-use
    def on_connect(self, client, userdata, flags, result_code):
        """ Callback when the MQTT client is connected.

        :param client: the client being connected.
        :param userdata: unused.
        :param flags: unused.
        :param result_code: result code.
        """
        debug_log("Connected with result code {}".format(result_code))
        self.state_handler.set_state(State.welcome)

    # pylint: disable=unused-argument
    def on_disconnect(self, client, userdata, result_code):
        """ Callback when the MQTT client is disconnected. In this case,
            the server waits five seconds before trying to reconnected.

        :param client: the client being disconnected.
        :param userdata: unused.
        :param result_code: result code.
        """
        debug_log("Disconnected with result code " + str(result_code))
        self.state_handler.set_state(State.goodbye)
        time.sleep(5)
        self.start()

    # pylint: disable=unused-argument
    def on_message(self, client, userdata, msg):
        """ Callback when the MQTT client received a new message.

        :param client: the MQTT client.
        :param userdata: unused.
        :param msg: the MQTT message.
        """
        if msg.payload is None or len(msg.payload) == 0:
            debug_log("New message on topic {}".format(msg.topic))
        if msg.topic == "hermes/nlu/intentParsed" and msg.payload:
            payload = json.loads(msg.payload.decode('utf-8'))
            intent = IntentParser.parse(payload, self.registry.intent_classes)
            debug_log("New intent: {}".format(str(intent.intentName)))
            if self.handle_intent:
                self.handle_intent(intent)
        elif msg.topic == "hermes/hotword/toggleOn":
            self.state_handler.set_state(State.hotword_toggle_on)
        elif msg.topic == "hermes/hotword/detected":
            if not self.first_hotword_detected:
                self.client.publish(
                    "hermes/feedback/sound/toggleOff", payload=None, qos=0, retain=False)
                self.first_hotword_detected = True
            else:
                self.state_handler.set_state(State.hotword_detected)
        elif msg.topic == "hermes/asr/toggleOn":
            self.state_handler.set_state(State.asr_toggle_on)
        elif msg.topic == "hermes/asr/textCaptured":
            self.state_handler.set_state(State.asr_text_captured)



# -*-: coding utf-8 -*-
""" Leds service for handling visual feedback for various states of
    the system. """

import random
import time
import struct
import threading

import respeaker.usb_hid as usb_hid

from .singleton import Singleton
from .thread_handler import ThreadHandler
from .usb_utils import USB


class LedsService:
    """ Leds service for handling visual feedback for various states of
        the system. """

    class State:
        """ Leds states.- """
        none, waking_up, standby, listening, loading, notify, error = range(7)

    def __init__(self, thread_handler):
        self.thread_handler = ThreadHandler()
        led_device = USB.get_boards()
        if led_device == USB.Device.respeaker:
            self.animator = ReSpeakerAnimator()
        else:
            self.animator = None

    def start_animation(self, animation_id):
        if not self.animator:
            return
        (animation, identifier) = self.get_animation(animation_id)
        
        self.thread_handler.run(target=self.animator.run, args=(identifier, animation, ))

    def get_animation(self, animation):
        identifier = str(random.randint(1, 100000))
        return (Animation(identifier, animation), identifier)


class Animation(Singleton):

    def __init__(self, id, active=0):
        self.id = id
        self.active = active


class ReSpeakerAnimator(object):

    def __init__(self):
        self.hid = usb_hid.get()
        self.led_dict = {
            'LED_1': 3,
            'LED_2': 4,
            'LED_3': 4,
            'LED_4': 5,
            'LED_5': 6,
            'LED_6': 7,
            'LED_7': 8,
            'LED_8': 9,
            'LED_9': 10,
            'LED_10': 11,
            'LED_11': 12,
            'LED_12': 13,
            'LED_13': 14,
        }

        self.animation_none = [
            {value: 0 for value in self.led_dict.values()}]
        self.animation_waking_up = [
            {value: 2 * st for value in self.led_dict.values()} for st in range(127)]
        self.animation_standby = [{value: st for value in self.led_dict.values(
        )} for st in list(range(31)) + list(reversed(range(31)))]

        # the matrix above:
        def value(i, j, N):
            if i == j:
                return 0
            elif i == (j + 1) % N:
                return 0
            elif i == (j - 1) % N:
                return 64
            elif i == (j + 2) % N or i == (j - 2) % N:
                return 127
            else:
                return 255

        self.animation_listening = []
        leds = range(3, 15)
        N = len(leds)

        for i in range(0, N):
            self.animation_listening.append(
                {leds[j]: value(i, j, N) for j in range(0, N)})

        self.animation_loading = [{value: 16 * st for value in self.led_dict.values()}
                                  for st in list(range(15)) + list(reversed(range(15)))]

        self.animation_notify = []
        for frame in range(0, 2):
            self.animation_notify.append(
                {value: int("0000ff", 16) for value in self.led_dict.values()})
            self.animation_notify.append(
                {value: int("000000", 16) for value in self.led_dict.values()})

        self.animation_error = []
        for frame in range(0, 3):
            self.animation_error.append(
                {value: int("ff0000", 16) for value in self.led_dict.values()})
            self.animation_error.append(
                {value: int("000000", 16) for value in self.led_dict.values()})

        # Set the ReSpeaker in the right mode
        try:
            init_addr = 0  # 0x0
            init_data = 6  # 0x00000006

            self.write(init_addr, init_data)
        except Exception as e:
            print(e.message)

    def write(self, address, data):
        if data > 0xFFFF:
            data = struct.pack('<I', data)
        elif data > 0xFF:
            data = struct.pack('<H', data)

        data = self.to_bytearray(data)
        length = len(data)
        if self.hid:
            packet = bytearray([address & 0xFF, (address >> 8) &
                                0x7F, length & 0xFF, (length >> 8) & 0xFF]) + data
            packet = list(packet)
            self.hid.write(packet)

    def read(self, address, length):
        if self.hid:
            self.hid.write(list(bytearray(
                [address & 0xFF, (address >> 8) & 0xFF | 0x80, length & 0xFF, (length >> 8) & 0xFF])))
            for _ in range(6):
                data = self.hid.read()
                # skip VAD data
                if int(data[0]) != 0xFF and int(data[1]) != 0xFF:
                    return data[4:(4 + length)]

    @staticmethod
    def to_bytearray(data):
        if type(data) is int:
            array = bytearray([data & 0xFF])
        elif type(data) is bytearray:
            array = data
        elif type(data) is str:
            array = bytearray(data)
        elif type(data) is list:
            array = bytearray(data)
        else:
            raise TypeError('%s is not supported' % type(data))

        return array

    def run(self, id, animation, run_event):
        if animation.active == LedsService.State.none:
            for item in self.animation_none:
                if animation.id != id or not run_event.is_set():
                    break
                for k, v in item.items():
                    self.write(k, v)
                time.sleep(0.5)
        elif animation.active == LedsService.State.waking_up:
            for item in self.animation_waking_up:
                if animation.id != id or not run_event.is_set():
                    break
                for k, v in item.items():
                    self.write(k, v)
                time.sleep(0.02)

        elif animation.active == LedsService.State.standby:
            while True:
                if animation.id != id or not run_event.is_set():
                    break
                for item in self.animation_standby:
                    if animation.id != id or not run_event.is_set():
                        break
                    for k, v in item.items():
                        self.write(k, v)
                    time.sleep(0.05)
        elif animation.active == LedsService.State.listening:
            while True:
                if animation.id != id or not run_event.is_set():
                    break
                for item in self.animation_listening:
                    if animation.id != id or not run_event.is_set():
                        break
                    for k, v in item.items():
                        self.write(k, v)
                    time.sleep(0.01)
        elif animation.active == LedsService.State.loading:
            while True:
                if animation.id != id or not run_event.is_set():
                    break
                for item in self.animation_loading:
                    if animation.id != id or not run_event.is_set():
                        break
                    for k, v in item.items():
                        self.write(k, v)
                    time.sleep(0.01)
        elif animation.active == LedsService.State.notify:
            for item in self.animation_notify:
                if animation.id != id or not run_event.is_set():
                    break
                for k, v in item.items():
                    self.write(k, v)
                time.sleep(0.1)
        elif animation.active == LedsService.State.error:
            for item in self.animation_error:
                if animation.id != id or not run_event.is_set():
                    break
                for k, v in item.items():
                    self.write(k, v)
                time.sleep(0.2)
